#!/usr/bin/env bash

# convertSizeToByte - converts a given size to byte
# parameter:
#  $1 - size in "kB", "kiB", "MB", "MiB", "GB", "GiB", "TB" or "TiB"
# return:
#  size in byte
convertSizeToByte() {
  local unit=$(echo $1 | sed "s/[0-9\.]*//g")
  local value=$(echo $1 | sed "s/[a-zA-Z%]*//g")
  local factor=1

  case $(echo ${unit} | tr [A-Z] [a-z]) in
    "b") factor=1 ;;
    "kb") factor=1000 ;;
    "mb") factor=1000000 ;;
    "gb") factor=1000000000 ;;
    "tb") factor=1000000000000 ;;
    "kib") factor=1024 ;;
    "mib") factor=1048576 ;;
    "gib") factor=1073741824 ;;
    "tib") factor=1099511627776 ;;
    *) factor=1 ;;
  esac

  res=$(dc -e "0k ${value} ${factor} * p")

  echo "${res%.*}"
}

# convertByteToSize - converts a byte size to a given unit 
# parameter:
#  $1 - size in byte
#  $2 - unit like "B", "kB", "kiB", "MB", "MiB", "GB", "GiB", "TB" or "TiB"
# return:
#  rounded size with unit
convertByteToSize() {
  local factor=1

  case $(echo $2 | tr [A-Z] [a-z]) in
    "b") factor=1 ;;
    "kb") factor=1000 ;;
    "mb") factor=1000000 ;;
    "gb") factor=1000000000 ;;
    "tb") factor=1000000000000 ;;
    "kib") factor=1024 ;;
    "mib") factor=1048576 ;;
    "gib") factor=1073741824 ;;
    "tib") factor=1099511627776 ;;
    *) factor=1 ;;
  esac

  res=$(dc -e "2k $1 ${factor} / p")
  echo "${res%.*}$2"
}

# validateDiskScheme - checks the scheme of a disk device
# parameter:
#  $1 - disk device name
# return:
#  0 - no differences
#  1 - current label differs from given label
validateScheme() {
  local lbl=$(echo ${diskdata[0]} | cut -f 6 -d ":")
  local retval=1
  local scheme=$(eval echo \$DISK_$(echo $1 | tr [a-z] [A-Z])_SCHEME | tr [A-Z] [a-z])

  if [ "x${scheme}" = "x${lbl}" ];
  then
    retval=0
  fi

  return ${retval}
}

# initialDiskScheme - initials the scheme of a disk device
# parameter:
#  $1 - disk device name
# return:
#  none
initialDiskScheme() {
  local scheme=$(eval echo \$DISK_$(echo $1 | tr [a-z] [A-Z])_SCHEME | tr [A-Z] [a-z])

  if [ ${scheme} = "mbr" -o ${scheme} = "msdos" ];
  then
    parted -s -m /dev/$1 mklabel msdos
  else
    parted -s -m /dev/$1 mklabel ${scheme}
  fi

  return
}

# validateDiskPartitionLayout - checks the partition layout of a disk device
# parameter:
#  $1 - disk device name
# return:
#  0 - no differences
#  1 - current partition layout differs from given partition layout 
validateDiskPartitionLayout() {
  declare -a parts=()
  local retval=0

  for part in $(eval "for p in \${DISK_$(echo $1 | tr [a-z] [A-Z])_PARTITIONS[@]}; do echo \$p; done ")
  do
    parts+=(${part})
  done

  if [ ${#parts[@]} -eq $((${#diskdata[@]} - 1)) ];
  then
    for idx in $(seq 1 1 ${#parts[@]})
    do
      gsize=$(convertSizeToByte $(echo ${parts[$((${idx} - 1))]} | cut -f 1 -d ";"))
      dsize=$(convertSizeToByte $(echo ${diskdata[${idx}]} | cut -f 4 -d ":"))
      diffperc=$(dc -e "2k $dsize $gsize - 100 * $gsize / p" | sed -e "s/^-//g")

      if [ "x${diffperc%.*}" = "x" ];
      then
        diffperc=0
      fi

      #if difference between current disk size and given disk size more than 1% then the partition layout has changed
      if [ ${diffperc%.*} -gt 1 ];
      then
        retval=1
      fi
    done
  else
    retval=1
  fi

  return ${retval}
}

# initialDiskPartitionLayout - initial the partition layout of a disk device
# parameter:
#  $1 - disk device name
# return:
#  none 
initialDiskPartitionLayout() {
  declare -a parts=()
  local scheme=$(eval echo \$DISK_$(echo $1 | tr [a-z] [A-Z])_SCHEME | tr [A-Z] [a-z])
  local start=$(convertSizeToByte "1MiB")
  local fs=""
  local size=""
  local end=""
  local idx=1

  initialDiskScheme $1

  # create partitions
  for part in $(eval "for p in \${DISK_$(echo $1 | tr [a-z] [A-Z])_PARTITIONS[@]}; do echo \$p; done ")
  do
    fs=$(echo ${part} | cut -f 2 -d ";")
    fsize=$(echo ${part} | cut -f 1 -d ";")
    size=$(convertSizeToByte $(echo ${part} | cut -f 1 -d ";"))
    end=$((${start} + ${size}))
    case ${scheme} in
      "mbr" | "msdos")
         parted -a optimal -s -m /dev/$1 mkpart primary ${fs} $(convertByteToSize ${start} "MiB") $(convertByteToSize ${end} "MiB")
         ;;

      "gpt")
         parted -a optimal -s -m /dev/$1 mkpart "Part${idx}" ${fs} $(convertByteToSize ${start} "MiB") $(convertByteToSize ${end} "MiB")
         ;;
    esac
    start=${end}
    idx=$(($idx + 1))
  done
}

# setDiskPartitionFlags - set the flags of a disk partition
# parameter:
#  $1 - disk device name
# return:
#  none
setDiskPartitionFlags() {
  local idx
  local flags

  # reset all flags of every partition
  for part in $(parted -s -m /dev/$1 print | sed -e "1,2d;s/[\ ;]//g" | cut -f 1,7 -d ":")
  do
    idx=$(echo ${part} | cut -f 1 -d ":")
    for flag in $(echo ${part} | sed "s/^[^:]*:\(.*\)$/\1/g;s/,/ /g")
    do    
      parted -s /dev/$1 set ${idx} ${flag} off
    done
  done

  # set all given flags of a partition
  idx=1
  for part in $(eval "for p in \${DISK_$(echo $1 | tr [a-z] [A-Z])_PARTITIONS[@]}; do echo \$p; done ")
  do
    flags=$(echo ${part} | cut -f 8 -d ";")
    if [ -n "${flags}" ];
    then
      for flag in $(echo ${flags} | sed "s/,/ /g")
      do
        parted -s /dev/$1 set ${idx} ${flag} on
      done
    fi
    idx=$(($idx + 1))
  done

  return
}

# getPartitionDeviceByIndex - returns the partition device of a disk by index
# parameter:
#  $1 - name of the disk device
#  $2 - index of the disk partition
# return:
#  0 - no error
#  1 - error occured
# hint:
#  The partition device of the given disk is returned on STDOUT.
getPartitionDeviceByIndex() {  
  local scheme=$(eval echo \$DISK_$(echo $1 | tr [a-z] [A-Z])_SCHEME | tr [A-Z] [a-z])
  local partdev=""
  local partname
  local retval=0

  if [ ${scheme} = "mbr" -o ${scheme} = "msdos" ];
  then
    partdev=$(fdisk -l /dev/$1 2>/dev/null | sed -n -e "s/^\(\/dev[^ ]*\) .*$/\1/p" | sed -n "$2p")
  else
    if [ -d /sys/block/$1 ]; 
    then
      partname=$(ls -d /sys/block/$1/$1* | sed -n "$2p" | xargs basename)
      if [ -n ${partname} -a -e /dev/${partname} ];
      then
        partdev="/dev/${partname}"
      else
        retval=1
      fi
    else
      retval=1
    fi
  fi

  echo ${partdev}

  return ${retval}
}  

# formatDiskPartion - format a partition
# parameter:
#  $1 - disk device name
#  $2 - disk partition index
#  $3 - filesystem name
# return:
#  0 - no error
#  1 - error occurred 
#  2 - unknown filesystem 
formatDiskPartion() {
  local device
  local retval=0

  device=$(getPartitionDeviceByIndex $1 $2)

  case $(echo $3 | tr "[A-Z]" "[a-z]") in
    "ext4") mkfs.ext4 -q -F ${device} >/dev/null|| retval=1 ;;
    "ext3") mkfs.ext3 -q -F ${device} >/dev/null || retval=1 ;;
    "fat32") mkfs.fat -F 32 ${device} >/dev/null || retval=1 ;;
    "fat16") mkfs.fat -F 16 ${device} >/dev/null || retval=1 ;;
    "swap") mkswap ${device} >/dev/null || retval=1 ;;
    "ntfs") mkntfs -q -f ${device} >/dev/null || retval=1 ;;
    "btrfs") mkfs.btrfs -q -f ${device} >/dev/null || retval=1 ;;
    "reiserfs") mkreiserfs -q ${device} >/dev/null || retval=1 ;;
    *) retval=2 ;;
  esac

  return ${retval}
}

# formatDiskPartitions - format partitions of a disk if necessary
# parameter:
#  $1 - disk device name
# return:
#  none
formatDiskPartitions() {
  declare -a parts=()
  local idx
  local gfs
  local dfs
  local format

  for part in $(eval "for p in \${DISK_$(echo $1 | tr [a-z] [A-Z])_PARTITIONS[@]}; do echo \$p; done ")
  do
    parts+=(${part})
  done

  for part in $(parted -s -m /dev/$1 print | sed -e "1,2d;s/[\ ;]//g" | cut -f 1,5 -d ";")
  do
    idx=$(echo ${part} | cut -f 1 -d ":")
    dfs=$(echo ${part} | cut -f 2 -d ":")
    gfs=$(echo ${parts[$((${idx} - 1))]} | cut -f 2 -d ";")
    format=$(echo ${parts[$((${idx} - 1))]} | cut -f 7 -d ";" | tr "[A-Z]" "[a-z]")

    if [ "${dfs}" != "${gfs}" -o "${format}" = "true" ];
    then
      formatDiskPartion $1 ${idx} ${gfs}
    fi
  done
}

# generateFSTab - generate fstab file
# parameter:
#  $1 - path of fstab file
#  $2 - mount with UUID
#  $3 - mount path prefix (optional)
# return:
#  none
generateFSTab() {
  local idx
  local uuid
  local blkidout
  local mntpnt
  local mntpntpfx=""
  local fs
  local opts
  local dmp
  local pass
  local dev
  local partdev

  if [ $# -eq 3 ];
  then
    mntpntpfx="$3"
  fi

  echo -e "#device           mount point           filesystem         options       dump    pass\n" >$1

  for disk in ${DISK_DEVICES[@]}
  do
    idx=1
    for part in $(eval "for p in \${DISK_$(echo ${disk} | tr [a-z] [A-Z])_PARTITIONS[@]}; do echo \$p; done ")
    do
      partdev=$(getPartitionDeviceByIndex ${disk} ${idx})
      uuid=$(blkid -s UUID -o value ${partdev})
      fs=$(blkid -s TYPE -o value ${partdev})
      mntpnt=$(echo ${part} | cut -f 3 -d ";") 
      opts=$(echo ${part} | cut -f 4 -d ";") 
      dmp=$(echo ${part} | cut -f 5 -d ";") 
      pass=$(echo ${part} | cut -f 6 -d ";") 
      
      if [ "${fs}" = "swap" -o "${fs}" = "linux-swap" ];
      then
        mntpnt="none"
      fi

      if [ -n "${fs}" -a -n "${mntpnt}" -a -n "${uuid}" ];
      then
         case ${2} in
          "uuid") dev="UUID=${uuid}" ;;
          *)
            dev=${partdev}
           ;;
         esac
         echo "#${dev}" >>$1
         if [ "${mntpnt}" != "none" ];
         then 
           echo -e "${dev}     ${mntpntpfx}${mntpnt}    ${fs}    ${opts}    ${dmp}    ${pass}\n" >>$1
         else
           echo -e "${dev}     ${mntpnt}    ${fs}    ${opts}    ${dmp}    ${pass}\n" >>$1
         fi
      fi
      idx=$((${idx} + 1))
    done
  done
}

# createTargetDirectoryLayout - create directory layout for target installation
# parameter:
#  $1 - path to target directory
# return:
#  none 
# hint:
#  This function mounts also the existing partitions.
createTargetDirectoryLayout(){
  local idx
  local mntpnt
  local fs
  local opts
  local partdev
  declare -a partmntdata=()

  # combine partition devices with their mount point and mount options 
  for disk in ${DISK_DEVICES[@]}
  do
    idx=1
    for part in $(eval "for p in \${DISK_$(echo ${disk} | tr [a-z] [A-Z])_PARTITIONS[@]}; do echo \$p; done ")
    do
      partdev=$(getPartitionDeviceByIndex ${disk} ${idx})
      fs=$(blkid -s TYPE -o value ${partdev})
      mntpnt=$(echo ${part} | cut -f 3 -d ";") 
      opts=$(echo ${part} | cut -f 4 -d ";") 
      if [ "${fs}" = "swap" -o "${fs}" = "linux-swap" ];
      then
        mntpnt="none"
      fi
      partmntdata+=("${partdev};${mntpnt};${fs};${opts}")
      idx=$((${idx} + 1))
    done
  done

  # create directory layout and mount partitions
  for part in $(eval "for p in \${partmntdata[@]}; do echo \$p; done" | sort -k 2 -t ";")
  do
    partdev=$(echo ${part} | cut -f 1 -d ";")   
    mntpnt=$(echo ${part} | cut -f 2 -d ";")   
    fs=$(echo ${part} | cut -f 3 -d ";")   
    opts=$(echo ${part} | cut -f 4 -d ";")   

    if [ "${mntpnt}" != "none" ];
    then
      mkdir -p $1/${mntpnt}
      if [ -z "${opts}" ];
      then
        mount -t ${fs} ${partdev} $1${mntpnt}
      else
        mount -t ${fs} ${partdev} $1${mntpnt} -o ${opts}
      fi 
    fi
  done
}

for disk in ${DISK_DEVICES[@]}
do
  declare -a diskdata=()

  for ddt in $(parted -s -m /dev/${disk} print 2>/dev/null | sed '1d;s/[ ]*//g')
  do
    diskdata+=(${ddt})
  done

  validateScheme ${disk}

  if [ $? -eq 0 ];
  then
    validateDiskPartitionLayout ${disk}

    if [ $? -ne 0 ];
    then
      initialDiskPartitionLayout ${disk}
    fi
 
  else
    initialDiskScheme ${disk}
    initialDiskPartitionLayout ${disk}
  fi  

  setDiskPartitionFlags ${disk}
  formatDiskPartitions ${disk}
done

generateFSTab "/tmp/fstab.local" "uuid"
createTargetDirectoryLayout "/target"
